/***************************************************************************
                          codegenerator.cpp  -  description
                             -------------------
    begin                : mer d�c 21 2005
    copyright            : (C) 2005 by Gibaud Alain
    email                : alain.gibaud@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "codegenerator.h"
#include "utils.h"

#include <iostream>
#include <sstream>
#include <algorithm>
#include <cstring>
using namespace std ;

// ====================================================
codeGenerator::codeGenerator(device *d)
{
    dev = d ;
}

codeGenerator::~codeGenerator()
{}

/** No descriptions */
string codeGenerator::processConfigBits(int wordsize, const string& configWordName, int offset)
{
    vector<deviceData *> cfbits  ;
    int prevBit = wordsize ;
    ostringstream ccode ;

    // 1 - isolate config bits
    cfbits = utils::filter( dev->data, "cfgbits") ;

    // 2 - search key "CONFIG"  (or "CONFIGxy", or CONFIG2)
    deviceData *conf = find( cfbits, "key", configWordName ) ;
    if( conf == 0 && configWordName != "CONFIG2")
    { // this register is not used , insert a filler
        ccode << "\n// " << configWordName << " ( Not defined for this device )"  << endl ;
        ccode << "\"0(" << (wordsize-1+offset) << "-" << offset << "){Unused\\n}\"\n" << endl ;
        return ccode.str() ;
    }
    else if(conf == 0)
    {  // missing CONFIG2 == only one config reg
        return "" ;
    }
    // 2.1 - search value of unused bits
    string unused_bits_str =  conf->params["unused"] ;
    unsigned int unused_bits ;
    utils::toUInt(unused_bits_str,unused_bits);


    string configaddr =  conf->params["addr"] ;

    // 3 - insulate "field" records
    vector<deviceData *> fields = utils::filter( conf->subdata, "field") ;
    splitNonContiguous( fields ) ;

    // sort config bits, by decreasing mask value
    sort( fields.begin(), fields.end(), utils::maskComparator ) ;

    ccode << "\n// " << configWordName << " (" << configaddr << ")" << endl ;

    vector<deviceData *>::iterator i ;
    for(i =  fields.begin()   ; i != fields.end() ; ++i)
    {
        string key = (*i)->params["key"] ;
        unsigned int mask ;
        utils::toUInt( (*i)->params["mask"] , mask) ;
        maskWarning(mask) ;
        string desc = (*i)->params["desc"] ;
        noSpuriousChars(desc) ;
        int bh, bl ;
        utils::maskToRange(mask, bh,bl) ;

        if( bh != prevBit -1)
        { // insert unused bits
            int zbh = prevBit-1, zbl = bh+1 ;
            ccode << "\"" << utils::bitState(zbl,zbh,unused_bits) << "(" ;
            if( zbh != zbl )
                ccode <<(zbh+offset)<<"-"<<(zbl+offset) ;
            else
                ccode << (zbl+offset) ;
            ccode << ")\""  ;
            ccode << "\n\"{Unused\\n}\"\n" ;
        }
        prevBit = bl ;

        ccode << "\"" <<key<<"[" ;
        if( bh != bl )
            ccode << (bh+offset) <<"-"<< (bl+offset) ;
        else
            ccode << (bl+offset) ;
        ccode << "]\"" ;

        ccode <<"\n\"{" << desc << "\\n\"" << endl << makeBitsDescription((*i)->subdata,bh, bl) << "\"}\"" << endl ;
    }

    if( prevBit != 0)
    {   // add filler to end of word
        int zbh = prevBit-1, zbl = 0 ;
        ccode << "\"" << utils::bitState(zbl,zbh,unused_bits) << "(" ;
        if( zbh != zbl )
            ccode <<(zbh+offset)<<"-"<<(zbl+offset) ;
        else
            ccode << (zbl+offset) ;
        ccode << ")\""  ;
        ccode << "\n\"{Unused\\n}\"\n" ;
    }

    ccode <<endl ;

    return ccode.str()   ;
}
/** No descriptions */
deviceData * codeGenerator::find(vector<deviceData *>& vec, const string& key, const string& value)
{
    vector<deviceData *>::iterator i ;
    for( i = vec.begin() ; i != vec.end() ; ++i)
    {
        map<string, string>& p = (*i)->params ;
        map<string, string>::iterator k = p.find(key) ;
        if( k != p.end() && (*k).second == value)
            return *i ;
    }
    return 0 ;
}
/** No descriptions */
string codeGenerator::makeBitsDescription(vector<deviceData *>& bits, int bh, int bl)
{
    vector<deviceData *>::iterator i ;
    string r = "\"" ;

    for(i =  bits.begin()   ; i != bits.end() ; ++i)
    {
        if( (*i)-> dataname == "setting")
        {
            string desc = (*i)->params["desc"] ;
            noSpuriousChars(desc) ;
            unsigned int value ;
            utils::toUInt((*i)->params["value"] , value)  ;
            r += utils::bitsToString(value, bh,bl) ;
            r += "=" ;
            r +=  desc ;
            r += "\\n"  ;
        }
    }

    r += "\"" ;
    return r ;
}
/** No descriptions */
string codeGenerator::processConfigRegs()
{
    string r ;

    int core  = dev->deviceCore(), wordsize = core   ;
    if( core == 16 )
        wordsize = 8 ;

    r += "\n// config bits descriptor\n," ;

    if( wordsize != 8 )
    {
        r += processConfigBits(wordsize, "CONFIG") ;
        // some devices have 2 config regs !!
        string r2 = processConfigBits(wordsize, "CONFIG2") ;
        if(r2 != "")
        {
            r += "\";\"\n" ;
            r += r2 ;
        }
    }
    else
    { // pic18 or dsPIC
        int nw ;
        string nn = "X" ;
        for(nw = 1 ; nw <= 7 ; ++nw)
        {
            nn[0] = (char)(nw+'0') ;
            r += processConfigBits(wordsize, "CONFIG"+nn+"H",8) ;
            r += processConfigBits(wordsize, "CONFIG"+nn+"L") ;
            if(nw != 7)
                r += "\";\"\n" ;
        }
    }
    return r ;
}
/** Remove or substitute delimters used by PiKdev config bits interpretor */
void codeGenerator::noSpuriousChars(string& s)
{
    unsigned int i ;

    for( i = 0 ; i < s.size() ; ++i)
    {
        if( s[i] == '{' )
            s[i] = '(' ;    // text delimiter
        else if ( s[i] == '}' )
            s[i] = ')' ;
        else if ( s[i] == ';' )
            s[i] = ',' ;  // 16 bit words delimiter
        // all other chars are allowed
    }
}
/** No descriptions */
string codeGenerator::processAdresses()
{
    // needed data
    string memaddr1, 	memaddr2 ;  ///< Limit indexes  of program adresses for this pic
    string idaddr1, 	idaddr2 ;     ///< Limit indexes of id data adresses  for this pic
    string eeaddr1 , 	eeaddr2 ;   ///< Limit indexes of EEprom  adresses  for this pic
    string cfaddr1,  cfaddr2 ;     ///< Indexes of config register(s)
    string calibration1, calibration2  ; ///< Limit Indexes of calibration words

    int core  = dev->deviceCore()   ;

    vector<deviceData *> data ;
    string s ,r ;

    // main programm memory
    memaddr1 = memaddr2 = "-1" ;
    data = utils::filter(dev->data,"pgmmem" ) ;
    if( data.size() != 0)
    {
        s = data[0]->params["region"] ;
        utils::splitRegion(s,memaddr1,memaddr2) ;
        if( core == 16)
        {
            memaddr1 += "/2" ;
            memaddr2 += "/2" ; // byte adresses
        }
    }
    r += ", " ;
    r += memaddr1 ;
    r += ", " ;
    r += memaddr2 ;
    r += " // program memory \n" ;

    // user ID
    idaddr1 = idaddr2 = "-1" ;
    data = utils::filter(dev->data,"userid" ) ;
    if( data.size() != 0)
    {
        s = data[0]->params["region"] ;
        utils::splitRegion(s,idaddr1,idaddr2) ;
        if( core == 16)
        {
            idaddr1 += "/2" ;
            idaddr2 += "/2" ; // byte adresses
        }
    }

    r += ", " ;
    r += idaddr1 ;
    r += ", " ;
    r += idaddr2 ;
    r += " // user ID data\n" ;

    // EEPROM  (hex file adresses are different than adresses of .dev files)
    eeaddr1 = eeaddr2 = "-1" ;
    data = utils::filter(dev->data,"eedata" ) ;
    if( data.size() != 0)
    {
        s = data[0]->params["region"] ;
        utils::splitRegion(s,eeaddr1,eeaddr2) ;
        if( core == 16)
        {
            eeaddr1 += "/2+0xF00000/2" ;
            eeaddr2 += "/2+0xF00000/2" ; // byte adresses
        }
        else
        {
            eeaddr1 += "+0x2100" ;
            eeaddr2 += "+0x2100" ;
        }
    }

    r += ", " ;
    r += eeaddr1 ;
    r += ", " ;
    r += eeaddr2 ;
    r += " // EEPROM data\n" ;

    // CONFIG registers
    cfaddr1 = cfaddr2 = "-1" ;
    data = utils::filter(dev->data,"cfgmem" ) ;
    if( data.size() != 0)
    {
        s = data[0]->params["region"] ;
        utils::splitRegion(s,cfaddr1,cfaddr2) ;
        if( core == 16)
        {
            cfaddr1 += "/2" ;
            cfaddr2 += "/2" ; // byte adresses
        }
        else
        {
            ;
        }
    }

    r += ", " ;
    r += cfaddr1 ;
    r += ", " ;
    r += cfaddr2 ;
    r += " // CONFIG data\n" ;

    // calibration word in program space
    calibration1 = calibration2 = "-1" ;

    // Attention, de nombreux fichiers sont faux (cette info est supprimee)
    data = utils::filter(dev->data,"calmem" ) ;
    if( data.size() != 0)
    {
        s = data[0]->params["region"] ;
        utils::splitRegion(s,calibration1, calibration2) ;
        // certains fichiers affichent une zone de calibration fantaisiste
        unsigned int ical1, ical2 ;
        if( utils::toUInt(calibration1, ical1) && utils::toUInt(calibration2, ical2) &&
                (((ical2 - ical1) > 4) || (ical1 > ical2) ) )
        {
            cout << "Warning: Calibration data seems buggy and have been disabled." << endl ;
	    cout << calibration1 << "  " << calibration2 << " ... decimal " << ical1 << " "<< ical2 << endl ; 
            calibration1 = calibration2 = "-1" ;
        }
        else if( core == 16)
        {
            calibration1 += "/2" ;
            calibration2 += "/2" ; // byte adresses
        }
    }

    r += ", " ;
    r += calibration1 ;
    r += ", " ;
    r += calibration2 ;
    r += " // calibration data words\n" ;
    return r ;
}
/** No descriptions */
string codeGenerator::processID()
{
    vector<deviceData *> devid ;
    unsigned int id = ~0U , mask = ~0U;
    string smask , sid , r = ", -1U // non autodetectable ";
    int core  = dev->deviceCore()   ;

    devid = utils::filter(dev->data,"devid" ) ;
    if( devid.size() != 0)
    {
        smask = devid[0]->params["idmask"] ;
        sid =  devid[0]->params["id"] ;

        if( utils::toUInt(smask, mask) && utils::toUInt(sid, id) )
        {
            id &= mask ;
            if( core == 16 )
                id |= 0x8000 ; // forge special 18Fxxx ID
            ostringstream rr ;
            rr << ", 0x" << hex << id << "  // device ID " << endl ;
            r = rr.str() ;
        }
    }
    else
    {
        cout << "No device ID" << endl ;
    }
    return r ;
}
/** No descriptions */
string codeGenerator::parseBuffersSize()
{
    vector<deviceData *> pgmdata ;
    int core  = dev->deviceCore()   ;
    string r = ", -1" ;

    pgmdata = utils::filter(dev->data,"pgming" ) ;
    if( pgmdata.size() != 0)
    {
        vector<deviceData *> latchesdata ;
        latchesdata = utils::filter(pgmdata[0]->subdata,"latches" ) ;
        if(  latchesdata.size() != 0)
        {
            string wrBufSize = latchesdata[0]->params["pgm"] ;
            if( wrBufSize != "")
                r = ", " + wrBufSize ;
            if(core == 16 && wrBufSize != "" )
                r += "/2" ;
        }
    }
    r += " // write buffer size \n" ;
    return r ;
}


void codeGenerator::maskWarning(unsigned int m)
{
    // counts numer of bit blocks
    int previous = 0 , bb , mask ;
    string bits ;

    for( bb =0, mask = 0x8000 ; mask ; mask >>=1)
    {
        bits += ('0' + ((m & mask) != 0)) ;
        if( previous == 0 && ( (m & mask) != 0 ) )
            ++bb ;
        previous = (m & mask) ;
    }

    if( bb == 0)
        cout << "Warning: nul config bit mask " << endl ;
    else if( bb > 1)
        cout << "Warning: non contiguous config bits : " << bits  << endl ;

}
/** Analyse masks for non contiguous field detection
and split relevants fields to several compacts sub-fields
 */
void codeGenerator::splitNonContiguous( vector<deviceData *>& fields)
{
    vector<deviceData *>::iterator i ;
    vector<deviceData *> newfields ;
    unsigned int mask ;

    for(i =  fields.begin()   ; i != fields.end() ; ++i)
    {
        utils::toUInt( (*i)->params["mask"] , mask) ;
        vector<unsigned int> submasks = subMasks(mask) ;
        if( submasks.size() > 1)
        {
            // modify first submask
            ostringstream smask ;
            smask << "0x" << hex << submasks[0] ;
            (*i)->params["mask"] = smask.str() ;
            // store other submasks
            for( unsigned int k = 1 ; k <  submasks.size() ; ++k)
            {
                newfields.push_back(new deviceData( **i )) ;
                ostringstream smask ;
                smask << "0x" << hex << submasks[k] ;
                (newfields.back())->params["mask"] = smask.str() ;
            }
        }
    }

    // now add new fields
    for( unsigned int k = 0 ; k < newfields.size() ; ++k)
    {
        fields.push_back(newfields[k]) ;
    }
}
/** Create a vector of compact masks from non compact one. */
vector<unsigned int> codeGenerator::subMasks(unsigned int m)
{
    // counts numer of bit blocks
    int previous = 0 ;
    unsigned int mask , submask  ;
    vector<unsigned int> r ;

    for(  mask = 0x8000, submask = 0 ; mask ; mask >>=1)
    {
        if(   (m & mask)  )
        {
            // add bit to submask
            submask |= mask ;
        }
        else  if( previous)
        {
            r.push_back(submask) ;
            submask = 0 ;
        }
        previous = (m & mask) ;
    }
    // flush current submask, if any
    if( submask )
        r.push_back(submask) ;

    return r ;
}
