/***************************************************************************
                          pikdbmaker.cpp  -  description
                             -------------------
    begin                : lun jan 2 2006
    copyright            : (C) 2006 by Gibaud Alain
    email                : alain.gibaud@free.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "pikdbmaker.h"

#include "device.h"
#include "deviceparser.h"
#include "codegenerator.h"
#include "utils.h"

#include "assert.h"

#include <fstream>
#include <string>
#include <iostream>
#include <sstream>
#include <cstring>
using namespace std ;

pikDBMaker::pikDBMaker(const string& en, bool v) : 
	verbose(v), editor_name(en) 
{
}
pikDBMaker::~pikDBMaker()
{
}
/** No descriptions */
string pikDBMaker::go( )
{
    char t[512] ;
    FILE *f ;
    vector<string> r ;
    string fname ;

    r.clear() ;
    // note: -1 (not -l)
    f = popen("ls -1 PIC10*.dev pic10*.dev PIC12*.dev pic12*.dev PIC16*.dev pic16*.dev PIC18*.dev pic18*.dev 2>&1","r") ;
    if( f == 0) return "Can't fork 'ls' command.";

    while(fgets(t,511,f) == t)
    {
        int i = strlen(t) - 1 ;
        if(i >= 0 && t[i] == '\n') t[i] = 0 ;
        fname = t ;
        // skip error messages from ls
        if( fname.find("ls:") == string::npos)
          r.push_back( fname );
    }
    if( pclose(f) == -1) return "Error while collecting .dev filename." ;
    // we now have a list of all device files

    ofstream out("device_data.hh") ;
    string  status ;
    if( !out.is_open() ) return "Can't create 'device_data.hh' file." ;

    out << "//" << endl ;
    out << "// This file is automatically generated by the 'pikdb' utility" << endl ;
    out << "// from Microchip .dev files. Do not edit."  << endl ;
    out << "// in case of errors, please contact the author : " << endl ;
    out << "//             alain.gibaud@free.fr" << endl ;
    out << "//" << endl ;
    
    out << "\ndeviceRecord devDB[] = { \n" ;

    status.clear() ;
    for( unsigned int i = 0 ; /* status.empty()  && */ i < r.size() ; ++i)
    {
        cout << "processing " <<  r[i] << endl ;
        status = makeRecord(r[i], out ) ;
        if( status != "")
        {
          cerr << status << endl ;
          f = popen((editor_name+" "+ r[i]).c_str(), "r") ;
          fclose(f) ;
          --i ;
        }
        else if( i != r.size()-1) out << ",\n" ;
    }
    out << "\n} ;\n" ;


    return "" ;
}
/** No descriptions */
string pikDBMaker::makeRecord(const string& filename, ostream& os)
{
    device dev ;
    deviceParser parser(&dev, verbose) ;
    codeGenerator cg(&dev) ;
    string stat  ;
    ostringstream CPPcode ;

    stat =  parser.parseDevice(filename) ;
    if( stat != "")
      return stat ;
    
    // dev.print(cout) ;

    // make data for this device
    CPPcode << "{\n" ;
    CPPcode << "\"" << utils::gpasmName(dev.name) <<  "\""  << endl ;
    CPPcode << cg.processID() ;
    CPPcode << cg.processConfigRegs() ;
    CPPcode << cg.processAdresses();
    CPPcode << cg.parseBuffersSize() ;
    CPPcode << "\n, " << dev.deviceCore() << " // bits per word " ;
    CPPcode << "\n}\n" ;

    os <<  CPPcode.str() ;

    return "" ;
}
